import inspect
from contextvars import ContextVar
from contextlib import asynccontextmanager, AbstractAsyncContextManager
from functools import wraps

from sqlalchemy.ext.asyncio import AsyncEngine, async_sessionmaker

__all__ = ['RepoBase']



class RepoBase(AbstractAsyncContextManager):
    def __init__(self, db: AsyncEngine):
        self.db = db
        self._asession_maker = async_sessionmaker(self.db, expire_on_commit=False)
        self._current_session: ContextVar = ContextVar('current_session', default=None)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        for name, method in cls.__dict__.items():
            if inspect.iscoroutinefunction(method) and not name.startswith('_'):
                def make_wrapper(m):
                    @wraps(m)
                    async def wrapper(self, *args, **kwargs):
                        async with self._session():
                            return await m(self, *args, **kwargs)
                    return wrapper
                setattr(cls, name, make_wrapper(method))

    @property
    def _s(self):
        return self._current_session.get()

    @asynccontextmanager
    async def _session(self):
        if self._current_session.get() is not None:
            yield
            return
        async with self._asession_maker.begin() as session:
            token = self._current_session.set(session)
            try:
                yield
            finally:
                self._current_session.reset(token)

    async def __aenter__(self):
        if self._current_session.get() is not None:
            raise RuntimeError("Nested transactions are not supported")
        session = self._asession_maker()
        try:
            await session.begin()
        except:
            await session.close()
            raise
        self._current_session.set(session)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        session = self._current_session.get()
        self._current_session.set(None)
        try:
            if exc_type is None:
                await session.commit()
            else:
                await session.rollback()
        finally:
            await session.close()



import dataclasses
import datetime
import warnings

from .cycler import AsyncCycler

__all__ = ['TimeWindow', 'HotCycler', 'HotAsyncCycler']



@dataclasses.dataclass(slots=True)
class TimeWindow:
    start: datetime.time
    end: datetime.time
    
    
    def contains(self, t: datetime.time) -> bool:
        if self.start <= self.end:  # normal window
            return self.start <= t <= self.end
        return t >= self.start or t <= self.end  # crosses midnight
    
    
    @staticmethod
    def start_duration(start: datetime.time, duration: datetime.timedelta) -> 'TimeWindow':
        dt = datetime.datetime.combine(datetime.date.today(), start)
        return TimeWindow(start, (dt + duration).time())
    
    
    @staticmethod
    def around_time(target: datetime.time, pre: datetime.timedelta, post: datetime.timedelta = None) -> 'TimeWindow':
        post = post or pre
        dt = datetime.datetime.combine(datetime.date.today(), target)
        return TimeWindow((dt - pre).time(), (dt + post).time())


class HotCycler(AsyncCycler):
    
    def __init__(self, name, func=None,
                 timetable: list[TimeWindow] = None,
                 HOT_SLEEP=0.2,
                 DEFAULT_SLEEP=1,
                 WAS_WORK_SLEEP=1,
                 ERROR_SLEEP=1 * 60):
        super().__init__(name, func,
                         DEFAULT_SLEEP=DEFAULT_SLEEP,
                         WAS_WORK_SLEEP=WAS_WORK_SLEEP,
                         ERROR_SLEEP=ERROR_SLEEP)
        self.timetable: list[TimeWindow] = sorted(timetable or [], key=lambda w: w.start)
        self.HOT_SLEEP = HOT_SLEEP
        self._window: TimeWindow | None = None  # cached ongoing window
    
    
    @property
    def current_window(self) -> TimeWindow | None:
        """Returns the ongoing hot window (cached) or the nearest upcoming one."""
        now_t = datetime.datetime.now().time()
        if self._window is not None and self._window.contains(now_t):
            return self._window
        self._window = None
        for w in self.timetable:
            if w.contains(now_t):
                self._window = w  # cache ongoing window
                return w
        for w in self.timetable:  # upcoming — not cached
            if w.start > now_t:
                return w
        return self.timetable[0] if self.timetable else None  # wrap around midnight
    
    
    def next_hot_start(self) -> datetime.datetime | None:
        w = self.current_window
        if w is None:
            return None
        now = datetime.datetime.now()
        if w.contains(now.time()):
            return None  # already hot
        return datetime.datetime.combine(now.date(), w.start)
    
    
    def get_sleep_time(self, was_work, was_error):
        if was_error:
            return self.ERROR_SLEEP
        w = self.current_window
        now = datetime.datetime.now()
        if w is None:
            return self.WAS_WORK_SLEEP if was_work else self.DEFAULT_SLEEP
        if w.contains(now.time()):
            return self.HOT_SLEEP
        wait = (datetime.datetime.combine(now.date(), w.start) - now).total_seconds()
        if wait < 0:
            wait += 86400  # window starts tomorrow
        return min(self.DEFAULT_SLEEP, max(0, wait))


class HotAsyncCycler(HotCycler):
    def __init__(self, *args, **kwargs):
        warnings.warn("HotAsyncCycler is deprecated, use HotCycler instead", DeprecationWarning, stacklevel=2)
        super().__init__(*args, **kwargs)



__all__ = ['mapl', 'filterl', 'mapc', 'filterc']


def mapl(f, it):
    return list(map(f, it))


def filterl(f, it):
    return list(filter(f, it))


def mapc(collect, f, it):
    return collect(map(f, it))


def filterc(collect, f, it):
    return collect(filter(f, it))



from .cycler import *
from .hot_cycler import *
from .utils import *
from .funcs import *
from .decor import *
from .repomixins import *
from .repobase import *
from .crudrepo import *
from .files import *
from .yamlfile import *
from . import testools



class DefaultRepo:
    
    def __init__(self):
        self.get_session = None
        self.get_async_session = None
    
    
    def get_all(self, entity_class, **kwargs):
        with self.get_async_session() as sess:
            return sess.query(entity_class).filter_by(**kwargs).all()
    
    
    def get_first(self, entity_class, id_=None, **kwargs):
        with self.get_async_session() as sess:
            return sess.query(entity_class).filter_by(**kwargs).first()
    
    
    def get_or_create(self, entity_class, id_=None, **kwargs):
        with self.get_async_session() as sess:
            obj = sess.query(entity_class).filter_by(**kwargs).first()
            if not obj:
                obj = entity_class(**kwargs)
                sess.add(obj)
            sess.commit()
            return obj
    
    
    def create_if_none(self, ormobj, **kwargs):
        with self.get_async_session() as sess:
            obj = sess.query(ormobj.__class__).filter_by(**kwargs).first()
            if not obj:
                obj = ormobj
                sess.add(obj)
            sess.commit()
            return obj
    
    
    def save(self, *args):
        with self.get_session() as sess:
            sess.add_all(args)
            sess.commit()
    
    
    def asave(self, *args):
        with self.get_async_session() as sess:
            sess.add_all(args)
            sess.commit()
    
    
    def delete(self, *args):
        with self.get_async_session() as sess:
            for a in args:
                sess.delete(a)
            sess.commit()



import json

import yaml

__all__ = ['read_yaml', 'read_json']



def read_yaml(path):
    with open(path, encoding="utf-8") as f:
        return yaml.safe_load(f)


def read_json(path):
    with open(path, encoding="utf-8") as f:
        return json.load(f)




import functools
import inspect
import logging
import time
import traceback
from contextvars import ContextVar

from .utils import to_async

__all__ = ['catch_print_swallow_exc', 'log_execution_time', 'log_call']


def catch_print_swallow_exc(func):
    if inspect.iscoroutinefunction(func):
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            try:
                return await func(*args, **kwargs)
            except:
                traceback.print_exc()
        return wrapper
    else:
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except:
                traceback.print_exc()
        return wrapper



def log_execution_time(func=None, *, level=logging.WARNING):
    def decorator(f):
        if inspect.iscoroutinefunction(f):
            @functools.wraps(f)
            async def wrapper(*args, **kwargs):
                start = time.perf_counter()
                try:
                    return await f(*args, **kwargs)
                except:
                    traceback.print_exc()
                finally:
                    logging.log(level, f"---- {f.__name__} executed in {time.perf_counter() - start:.6f} seconds")
            return wrapper
        else:
            @functools.wraps(f)
            def wrapper(*args, **kwargs):
                start = time.perf_counter()
                try:
                    return f(*args, **kwargs)
                except:
                    traceback.print_exc()
                finally:
                    logging.log(level, f"---- {f.__name__} executed in {time.perf_counter() - start:.6f} seconds")
            return wrapper

    if func is not None:
        return decorator(func)
    return decorator



_call_indent: ContextVar[int] = ContextVar('call_indent', default=0)


def log_call(level=logging.INFO, logger=logging.getLogger()):
    def actual_decorator(func):
        if inspect.iscoroutinefunction(func):
            @functools.wraps(func)
            async def wrapper(*args, **kwargs):
                indent = _call_indent.get()
                i = ' ' * (indent * 8)
                token = _call_indent.set(indent + 1)
                logger.log(level, f"{i}{func.__name__} Start")
                try:
                    res = await func(*args, **kwargs)
                    logger.log(level, f"{i}{func.__name__} Finish")
                    return res
                except:
                    logger.error(f"{i}{func.__name__} Error")
                    raise
                finally:
                    _call_indent.reset(token)
            return wrapper
        else:
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                indent = _call_indent.get()
                i = ' ' * (indent * 8)
                token = _call_indent.set(indent + 1)
                logger.log(level, f"{i}{func.__name__} Start")
                try:
                    res = func(*args, **kwargs)
                    logger.log(level, f"{i}{func.__name__} Finish")
                    return res
                except:
                    logger.error(f"{i}{func.__name__} Error")
                    raise
                finally:
                    _call_indent.reset(token)
            return wrapper

    return actual_decorator



import yaml
import yamlordereddictloader

__all__ = ['YamlFile']



class YamlFile:
    
    def __init__(self, path):
        self.filepath = path
        self.d = None
    
    
    def read(self):
        with open(self.filepath, 'r', encoding="utf-8") as file:
            self.d = yaml.load(file, yamlordereddictloader.Loader)
        # new_dict = OrderedDict(list({'id': i}.items()) + list(q.items()))
    
    
    def save(self):
        with open(self.filepath, 'w', encoding="utf-8") as file:
            yaml.dump(self.d, file, yamlordereddictloader.Dumper, allow_unicode=True, sort_keys=False)




import datetime
import uuid
from typing import Annotated

__all__ = [
    'auto_repr', 'utcnow', 'undashed_uuid',
    'intpk', 'strpk', 'uuidpk', 'created_col', 'modified_col',
    'SerialPK_Mixin', 'UUIDPK_Mixin', 'TimestampMixin',
]

from sqlalchemy import DateTime, Integer, String
from sqlalchemy.orm import Mapped, mapped_column
from sqlalchemy.sql import func



def auto_repr(cls):
    def __repr__(self):
        columns = self.__table__.columns.keys()
        values = {col: getattr(self, col) for col in columns}
        return f"{self.__class__.__name__}({', '.join(f'{k}={v!r:.10}' for k, v in values.items())})"
    
    
    cls.__repr__ = __repr__
    return cls


def utcnow():
    return datetime.datetime.now(datetime.UTC)


def undashed_uuid() -> str:
    return uuid.uuid4().hex.upper()



intpk = Annotated[int, mapped_column(primary_key=True)]
strpk = Annotated[str, mapped_column(primary_key=True)]
uuidpk = Annotated[str, mapped_column(primary_key=True, default=undashed_uuid)]
created_col = Annotated[datetime.datetime, mapped_column(DateTime(timezone=True), server_default=func.now(), sort_order=1000)]
modified_col = Annotated[datetime.datetime, mapped_column(DateTime(timezone=True), server_default=func.now(), onupdate=utcnow, sort_order=1001)]


class SerialPK_Mixin:
    id = mapped_column(Integer, sort_order=-1000, primary_key=True)


class UUIDPK_Mixin:
    uuid = mapped_column(String, primary_key=True, default=undashed_uuid, sort_order=-999)


class TimestampMixin:
    created: Mapped[created_col]
    modified: Mapped[modified_col]






import asyncio
import datetime
import logging
from typing import Any

__all__ = ['AsyncCycler', 'SyncCycler']


class AsyncCycler:
    IDLE_LOG_EVERY = 10


    def __init__(self, name, async_cycled_func=None, DEFAULT_SLEEP=1, WAS_WORK_SLEEP=1, ERROR_SLEEP=1 * 60):
        self.name = name
        if async_cycled_func is not None:
            self.async_cycled_func = async_cycled_func
        self.DEFAULT_SLEEP = DEFAULT_SLEEP
        self.WAS_WORK_SLEEP = WAS_WORK_SLEEP
        self.ERROR_SLEEP = ERROR_SLEEP
        self.cycle_count: int = 0
        self.idle_count: int = 0
        self.last_cycle_at: datetime.datetime | None = None
        self.last_error_at: datetime.datetime | None = None


    def get_sleep_time(self, was_work, was_error):
        if was_error:
            return self.ERROR_SLEEP
        if was_work:
            return self.WAS_WORK_SLEEP
        return self.DEFAULT_SLEEP


    def health(self) -> dict:
        return {
            "name"         : self.name,
            "cycle_count"  : self.cycle_count,
            "idle_count"   : self.idle_count,
            "last_cycle_at": self.last_cycle_at.isoformat() if self.last_cycle_at else None,
            "last_error_at": self.last_error_at.isoformat() if self.last_error_at else None,
        }


    async def run(self):
        while True:
            was_work = False
            was_error = False
            try:
                was_work = await self.async_cycled_func()
            except Exception:
                was_error = True
                self.last_error_at = datetime.datetime.now()
                logging.exception(f"{self.name} | cycle error | sleep={self.ERROR_SLEEP}s")
            finally:
                sleep_time = self.get_sleep_time(was_work, was_error)
            self._update_stats(was_work)
            if was_work:
                logging.info(f"{self.name} | cycle={self.cycle_count} completed | sleep={sleep_time}s")
            elif not was_error and self.idle_count % self.IDLE_LOG_EVERY == 0:
                logging.debug(f"{self.name} | heartbeat | idle={self.idle_count}")
            await asyncio.sleep(sleep_time)


    def _update_stats(self, was_work: bool | Any):
        if was_work:
            self.cycle_count += 1
            self.idle_count = 0
            self.last_cycle_at = datetime.datetime.now()
        else:
            self.idle_count += 1


    async def async_cycled_func(self):
        return False


class SyncCycler(AsyncCycler):
    def cycled_func(self):
        return False


    async def async_cycled_func(self):
        return self.cycled_func()


    def run(self):
        asyncio.run(super().run())



from sqlalchemy import select
from .repobase import RepoBase

__all__ = ['CRUDRepo']



class CRUDRepo(RepoBase):
    
    async def get_all(self, entity_class, **kwargs):
        stmt = select(entity_class).filter_by(**kwargs)
        result = await self._s.execute(stmt)
        return result.scalars().all()
    
    
    async def get_first(self, entity_class, **kwargs):
        stmt = select(entity_class).filter_by(**kwargs)
        result = await self._s.execute(stmt)
        return result.scalars().first()
    
    
    async def create_if_none(self, ormobj, **kwargs):
        stmt = select(ormobj.__class__).filter_by(**kwargs)
        result = await self._s.execute(stmt)
        obj = result.scalars().first()
        if not obj:
            obj = ormobj
            self._s.add(obj)
        return obj
    
    
    async def create(self, *args):
        for arg in args:
            await self._s.merge(arg)
    
    
    async def update(self, *args):
        for arg in args:
            await self._s.merge(arg)
    
    
    async def upsert(self, *args):
        for arg in args:
            await self._s.merge(arg)
    
    
    async def delete(self, *args):
        for arg in args:
            self._s.delete(arg)



import asyncio
import inspect

from datetime import time

__all__ = [
    'get_or_first', 'first_or_none', 'map_many', 'applyer',
    'reverse_dict', 'merge', 'to_async', 'normalize_time', 'RAISE_IF_NONE',
]



def _Raise_if_len0(lst, err_text):
    if len(lst) == 0:
        raise Exception(err_text)


def RAISE_IF_NONE(var, err_text):
    if var is None:
        raise Exception(err_text)


def get_or_first(d, k=None):
    if k is None:
        return next(iter(d.values()))
    return d.get(k, next(iter(d.values())))


def first_or_none(a):
    return a[0] if len(a) > 0 else None


def map_many(iterable, function, *other):
    if other:
        return map_many(map(function, iterable), *other)
    return map(function, iterable)


def applyer(*funcs):
    def _applyer(x):
        for func in funcs:
            x = func(x)
        return x
    
    
    return _applyer


from collections import defaultdict



def reverse_dict(d):
    reversed_d = defaultdict(list)
    
    for key, value in d.items():
        if isinstance(value, list):
            for v in value:
                reversed_d[v].append(key)
        else:
            reversed_d[value].append(key)
    
    return {k: v[0] if len(v) == 1 else v for k, v in reversed_d.items()}


def merge(D1s: dict, *D2s: dict):
    def merge_two(D1: dict, D2: dict):
        for key, value in D1.items():
            if key in D2:
                if type(value) is dict and type(D2[key]) is dict:
                    merge(D1[key], D2[key])
                else:
                    if not isinstance(value, list):
                        D1[key] = [value]
                    if type(D2[key]) is list:
                        D1[key].extend(D2[key])
                    else:
                        D1[key].append(D2[key])
        for key, value in D2.items():
            if key not in D1:
                D1[key] = [value]
    
    
    for D2n in D2s:
        merge_two(D1s, D2n)


async def to_async(func, *args, **kwargs):
    if inspect.iscoroutinefunction(func):
        return await func(*args, **kwargs)
    else:
        return func(*args, **kwargs)


def normalize_time(hour: int, minute: int) -> time:
    hour += minute // 60
    minute %= 60
    hour %= 24
    return time(hour, minute)



def tracking_setattr(monkeypatch, history_list, cls, attr_name):
    original_setter = cls.__setattr__
    
    
    def tracking_setattr(self, name, value):
        if name == attr_name:
            history_list.append(value)
        original_setter(self, name, value)
    
    
    monkeypatch.setattr(cls, '__setattr__', tracking_setattr)


class AttrSpy:
    def __init__(self, initial_value=None):
        self.value = initial_value
        self.history = [initial_value]
    
    
    def __set__(self, instance, value):
        self.value = value
        self.history.append(value)
    
    
    def __get__(self, instance, owner):
        return self.value
